""" Circuits. 

Inspired by a mix of Cirq and Qiskit circuits.
"""

import functools
from flax import struct
from enum import Enum
from jax import Array, config
from typing import List, Optional, Union
from math import prod
from copy import deepcopy
from numbers import Number
from numpy import argsort
import jax.numpy as jnp
import jax.scipy as jsp

from jaxquantum.core.settings import SETTINGS
from jaxquantum.core.qarray import Qarray
from jaxquantum.core.operators import identity
from jaxquantum.circuits.gates import Gate
from jaxquantum.circuits.constants import SimulateMode

config.update("jax_enable_x64", True)

@struct.dataclass
class Register:
    dims: List[int] = struct.field(pytree_node=False)

    @classmethod
    def create(
        cls, dims: List[int]
    ):
        return Register(
            dims = dims
        )

    def __eq__(
        self, other
    ):
        if isinstance(other, Register):
            return self.dims == other.dims
        return False

@struct.dataclass
class Operation:
    gate: Gate
    indices: List[int] = struct.field(pytree_node=False)
    register: Register

    @classmethod
    def create(
        cls,
        gate: Gate,
        indices: Union[int,List[int]],
        register: Register
    ):
        if isinstance(indices, int):
            indices = [indices]

        assert gate.num_modes == len(indices), "Number of indices must match gate's num_modes."
        assert gate.dims == [register.dims[ind] for ind in indices], "Indices must match register dimensions."

        if any(
            (0 > ind and ind >= len(register.dims)) or not isinstance(ind, int) for ind in indices):
            raise ValueError("Indices must be integers within the register.")

        return Operation(
            gate = gate,
            indices = indices,
            register = register
        )

@struct.dataclass
class Layer:
    operations: List[Operation] = struct.field(pytree_node=False)
    _unique_indices: List[int] = struct.field(pytree_node=False)
    _default_simulate_mode: SimulateMode = struct.field(pytree_node=False)

    @classmethod
    def create(
        cls, operations: List[Operation], default_simulate_mode=SimulateMode.UNITARY
    ):
        all_indices = [ind for op in operations for ind in op.indices]
        unique_indices = list(set(all_indices))
        if len(all_indices) != len(unique_indices):
            raise ValueError("Operations must not have overlapping indices.")
        return Layer(
            operations = operations,
            _unique_indices = unique_indices,
            _default_simulate_mode = default_simulate_mode
        )

    def add(self, operation: Operation):
        if any(ind in self._unique_indices for ind in operation.indices):
            raise ValueError("Operations must not have overlapping indices.")
        self.operations.append(operation)
        self._unique_indices.extend(operation.indices)

    def gen_U(self):
        U = None

        if len(self.operations) == 0:
            return U
        
        indices_order = []
        for operation in self.operations:
            indices_order += operation.indices

            if U is None:
                U = operation.gate.U
            else:
                U = U ^ operation.gate.U
            
        register = self.operations[0].register
        missing_indices = [i for i in range(len(register.dims)) if i not in indices_order]

        for j in missing_indices:
            U = U ^ identity(register.dims[j])

        combined_indices = indices_order + missing_indices

        sorted_ind = list(argsort(combined_indices))
        U = U.transpose(sorted_ind)
        return U

    def gen_KM(self):
        KM = []

        if len(self.operations) == 0:
            return KM

        
        indices_order = []
        for operation in self.operations:
            indices_order += operation.indices

            if len(KM) == 0:
                KM = deepcopy(operation.gate.KM)
            else:
                # updated_KM = []                
                # for op1 in KM:
                #     for op2 in operation.gate.KM:
                #         updated_KM.append(op1^op2)
                # KM = updated_KM
                KM = KM.arraytensor(operation.gate.KM)
        
        register = self.operations[0].register
        missing_indices = [i for i in range(len(register.dims)) if i not in indices_order]
        
        for j in missing_indices:
            KM = KM ^ identity(register.dims[j])
        
        combined_indices = indices_order + missing_indices
        sorted_ind = list(argsort(combined_indices))
                
        KM = KM.transpose(sorted_ind)

        return KM


@struct.dataclass
class Circuit:
    register: Register
    layers: List[Layer]  = struct.field(pytree_node=False)
    

    @classmethod
    def create(
        cls, register: Register, layers: Optional[List[Layer]] = None
    ):  
        if layers is None:
            layers = []

        return Circuit(
            register = register,
            layers = layers,   
        )

    def append_layer(self, layer: Layer):
        self.layers.append(layer)

    def append_operation(self, operation: Operation, default_simulate_mode=SimulateMode.UNITARY):
        assert operation.register == self.register, f"Mismatch in operation register {operation.register} and circuit register {circuit.register}."
        self.append_layer(Layer.create([operation], default_simulate_mode=default_simulate_mode))

    def append(self, gate: Gate, indices: Union[int, List[int]], default_simulate_mode=SimulateMode.UNITARY):
        operation = Operation.create(gate, indices, self.register)
        self.append_operation(operation, default_simulate_mode=default_simulate_mode)


import pytest

import sys
import os

# Add the jaxquantum directory to the sys.path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import jaxquantum as jqt
import jax.numpy as jnp

minimum_version_for_tests = "0.2.0"


# Version
# ========================================
def test_version():
    high_enough = False 
    ver_tuple = tuple(map(int, jqt.__version__.split('.')))
    test_ver_tuple = tuple(map(int, minimum_version_for_tests.split('.')))
    high_enough = high_enough or (ver_tuple[0] > test_ver_tuple[0])
    high_enough = high_enough or (ver_tuple[0] == test_ver_tuple[0] and ver_tuple[1] > test_ver_tuple[1])
    high_enough = high_enough or (ver_tuple[0] == test_ver_tuple[0] and ver_tuple[1] == test_ver_tuple[1] and ver_tuple[2] >= test_ver_tuple[2])
    assert high_enough
# ========================================

# Initialization
# ========================================
def test_qarray_creation():
    a = jqt.Qarray.create(jnp.array([1,2,3]))
    assert a.shape == (3,1)

    a = jqt.Qarray.create(jnp.array([[1,2,3],[4,5,6]]))
    assert a.shape == (2,3,1)

    a = jqt.Qarray.create(jnp.array([[1,2,],[4,5]]), bdims=(2,))
    assert a.shape == (2,2,1)

    a = jqt.Qarray.from_list([])
    assert a.dims == ((),()) and a.shape == jnp.array([]).shape

    a = jqt.displace(3,1.0)
    b = jqt.displace(3,2.0)
    assert jqt.Qarray.from_array(a) == a
    assert jqt.Qarray.from_array([[[a,a]],[[b,b]]])[1][0][0] == b
    # assert jqt.Qarray.from_array([[],[]])

# =========================================

# Properties
# ========================================
def test_qarray_properties():
    a = jqt.Qarray.create(jnp.array([[1+1.0j,2,3],[4,5,6]]))
    assert a.qtype == jqt.Qtypes.ket
    assert a.dtype == jnp.array([1+1.0j]).dtype
    assert a.bdims == (2,)
    assert a.qdims == jqt.basis(3,1).qdims

    a_dag = a.dag()
    assert a.space_dims == a.dims[0]
    assert a_dag.space_dims == a_dag.dims[1]
    
    # a_bad = a.copy()
    # a_bad._qdims._qtype = None
    # with pytest.raises(ValueError): # Doesn't give expected error for some reason...
    #     print(a_bad.space_dims) 
    
    assert jnp.all(a.shaped_data == a.data.reshape(2,3,1))

    assert a.is_batched == True
    assert a[0] == jqt.Qarray.create(jnp.array([1+1.0j,2,3]))
    
    with pytest.raises(ValueError):
        print(a[0][0])

    a_reshaped = a.reshape_bdims(2,1)
    assert a_reshaped.shape == (2,1,3,1)

    assert len(a) == 2
    with pytest.raises(ValueError):
        print(len(a[0]))
    
    with pytest.raises(ValueError):
        print(a == a.data)

    assert a != a[0]
    assert a != jqt.Qarray.create(jnp.array([1,2]))

    with pytest.raises(ValueError):
        print(a/a_dag)

    assert a/3 == 1/3*a
    

# Basic Math
# ========================================
def test_qarray_basic_math_add():
    N = 3

    a = jqt.displace(N,1.0)
    b = jqt.displace(N,1.25)
    c = jqt.displace(N,1.5)
    
    scalar = 1.23
    scalar_id_data = scalar*jnp.eye(N)

    arr = jqt.Qarray.from_list([a,b])

    # Scalar + Qarray 
    assert jnp.max(jnp.abs(((scalar+a).data) - (scalar_id_data + a.data))) < 1e-10

    # Qarray + Scalar
    assert jnp.max(jnp.abs(((a+scalar).data) - (a.data + scalar_id_data))) < 1e-10

    # Scalar + QarrayArray
    assert jnp.max(jnp.abs(((scalar + arr).data - (scalar_id_data + arr.data)))) < 1e-10

    # QarrayArray + Scalar
    assert jnp.max(jnp.abs(((arr + scalar).data - (arr.data + scalar_id_data))) < 1e-10)


    tarr = jnp.array([[1,2,3,4],[4,5,6,7.2]])
    tarr_id = tarr.reshape(*tarr.shape,1,1) * jnp.eye(N)
    tqarr = jqt.displace(N,tarr)

    # Array + Qarray
    assert jnp.max(jnp.abs((tarr+b).data - (tarr_id + b.data))) < 1e-10

    # Qarray + Array
    assert jnp.max(jnp.abs((a+tarr).data - (a.data + tarr_id)) < 1e-10)

    # QarrayArray + Array 
    assert jnp.max(jnp.abs((tqarr + tarr).data - (tqarr.data + tarr_id))) < 1e-10

    # Array + QarrayArray
    assert jnp.max(jnp.abs((tarr + tqarr).data - (tarr_id + tqarr.data))) < 1e-10

    # QarrayArray + Qarray
    assert jnp.max(jnp.abs((arr + c).data - (arr.data + c.data))) < 1e-10
    
    # Qarray + QarrayArray
    assert jnp.max(jnp.abs((c + arr).data - (c.data + arr.data))) < 1e-10

    # Qarray + Qarray
    assert jnp.max(jnp.abs((a+b).data - (a.data + b.data))) < 1e-10
    
    # QarrayArray + QarrayArray
    assert jnp.max(jnp.abs((arr + arr).data - (arr.data + arr.data))) < 1e-10

    # QarrayArray + QarrayArray (of different size)
    with pytest.raises(TypeError):
        jqt.Qarray.from_list([a,b]) + jqt.Qarray.from_list([a,b,c])

def test_qarray_basic_math_sub():
    N = 3

    a = jqt.displace(N,1.0)
    b = jqt.displace(N,1.25)
    c = jqt.displace(N,1.5)
    
    scalar = 1.23
    scalar_id_data = scalar*jnp.eye(N)

    arr = jqt.Qarray.from_list([a,b])

    # Scalar - Qarray 
    assert jnp.max(jnp.abs(((scalar-a).data) - (scalar_id_data - a.data))) < 1e-10

    # Qarray - Scalar
    assert jnp.max(jnp.abs(((a-scalar).data) - (a.data - scalar_id_data))) < 1e-10

    # Scalar - QarrayArray
    assert jnp.max(jnp.abs((scalar - arr).data - (scalar_id_data - arr.data))) < 1e-10

    # QarrayArray - Scalar
    assert jnp.max(jnp.abs((arr - scalar).data - (arr.data - scalar_id_data))) < 1e-10


    tarr = jnp.array([[1,2,3,4],[4,5,6,7.2]])
    tarr_id = tarr.reshape(*tarr.shape,1,1) * jnp.eye(N)
    tqarr = jqt.displace(N,tarr)

    # Array - Qarray
    assert jnp.max(jnp.abs((tarr - b).data - (tarr_id - b.data))) < 1e-10

    # Qarray - Array
    assert jnp.max(jnp.abs((a-tarr).data - (a.data - tarr_id))) < 1e-10

    # QarrayArray - Array 
    assert jnp.max(jnp.abs((tqarr - tarr).data - (tqarr.data - tarr_id))) < 1e-10

    # Array - QarrayArray
    assert jnp.max(jnp.abs((tarr - tqarr).data - (tarr_id - tqarr.data))) < 1e-10

    # QarrayArray - Qarray
    assert jnp.max(jnp.abs((arr - c).data - (arr.data - c.data))) < 1e-10
    
    # Qarray - QarrayArray
    assert jnp.max(jnp.abs((c - arr).data - (c.data - arr.data))) < 1e-10

    # Qarray - Qarray
    assert jnp.max(jnp.abs((a-b).data - (a.data - b.data))) < 1e-10

    # QarrayArray - QarrayArray
    assert jnp.max(jnp.abs((arr - arr).data - (arr.data - arr.data))) < 1e-10

    # QarrayArray - QarrayArray (of different size)
    with pytest.raises(TypeError):
        jqt.Qarray.from_list([a,b]) - jqt.Qarray.from_list([a,b,c])

def test_qarray_basic_math_mul():
    N = 3

    a = jqt.displace(N,1.0)
    b = jqt.displace(N,1.25)
    c = jqt.displace(N,1.5)
    
    scalar = 1.23
    scalar_id_data = scalar*jnp.eye(N)

    arr = jqt.Qarray.from_list([a,b])

    # Scalar * Qarray 
    assert jnp.max(jnp.abs(((scalar*a).data) - (scalar * a.data))) < 1e-10

    # Qarray * Scalar
    assert jnp.max(jnp.abs(((a*scalar).data) - (a.data * scalar))) < 1e-10

    # Scalar * QarrayArray
    assert jnp.max(jnp.abs((scalar * arr).data - (scalar * arr.data))) < 1e-10

    # QarrayArray * Scalar
    assert jnp.max(jnp.abs((arr * scalar).data - (arr.data * scalar))) < 1e-10

    # ----

    tarr = jnp.array([[1,2,3,4],[4,5,6,7.2]])
    tarr_expanded = tarr.reshape(*tarr.shape,1,1)
    tqarr = jqt.displace(N,tarr)

    # Array * Qarray
    assert jnp.max(jnp.abs((tarr * b).data - (tarr_expanded * b.data))) < 1e-10

    # Qarray * Array
    assert jnp.max(jnp.abs((a * tarr).data - (a.data * tarr_expanded))) < 1e-10

    # QarrayArray * Array 
    assert jnp.max(jnp.abs((tqarr * tarr).data - (tqarr.data * tarr_expanded))) < 1e-10

    # Array * QarrayArray
    assert jnp.max(jnp.abs((tarr * tqarr).data - (tarr_expanded * tqarr.data))) < 1e-10

    # ----

    # QarrayArray * Qarray
    assert jnp.max(jnp.abs((arr * c).data - (arr.data @ c.data))) < 1e-10
    
    # Qarray * QarrayArray
    assert jnp.max(jnp.abs((c * arr).data - (c.data @ arr.data))) < 1e-10

    # Qarray * Qarray
    assert jnp.max(jnp.abs((a * b).data - (a.data @ b.data))) < 1e-10

    # QarrayArray * QarrayArray
    assert jnp.max(jnp.abs((arr * arr).data - jnp.einsum('nij,njk->nik', arr.data, arr.data))) < 1e-10

    # QarrayArray * QarrayArray (of different size)
    with pytest.raises(ValueError):
        jqt.Qarray.from_list([a,b]) @ jqt.Qarray.from_list([a,b,c])

def test_qarray_basic_math_matmul():
    N = 3

    a = jqt.displace(N,1.0)
    b = jqt.displace(N,1.25)
    c = jqt.displace(N,1.5)
    
    arr = jqt.Qarray.from_list([a,b])

    # ----

    tarr = jnp.array([[1,2,3,4],[4,5,6,7.2]])
    tqarr = jqt.displace(N,tarr)

    # Array @ Qarray
    with pytest.raises(TypeError):
        tarr @ b

    # Qarray @ Array
    with pytest.raises(TypeError):
        a @ tarr
    # assert jnp.max(jnp.abs((a @ b.data).data - (a.data @ b.data))) < 1e-10

    # QarrayArray @ Array 
    with pytest.raises(TypeError):
        tqarr @ tarr
    # assert jnp.max(jnp.abs((tqarr @ tarr).data - (tqarr.data @ c.data))) < 1e-10

    # Array @ QarrayArray
    with pytest.raises(TypeError):
        tarr @ tqarr
    # assert jnp.max(jnp.abs((tarr @ tqarr).data - (c.data @ tqarr.data))) < 1e-10

    # ----

    # QarrayArray @ Qarray
    assert jnp.max(jnp.abs((arr @ c).data - (arr.data @ c.data))) < 1e-10
    
    # Qarray @ QarrayArray
    assert jnp.max(jnp.abs((c @ arr).data - (c.data @ arr.data))) < 1e-10

    # Qarray @ Qarray
    assert jnp.max(jnp.abs((a @ b).data - (a.data @ b.data))) < 1e-10

    # QarrayArray @ QarrayArray
    assert jnp.max(jnp.abs((arr @ arr).data - jnp.einsum('nij,njk->nik', arr.data, arr.data))) < 1e-10

    # QarrayArray @ QarrayArray (of different size)
    with pytest.raises(ValueError):
        jqt.Qarray.from_list([a,b]) @ jqt.Qarray.from_list([a,b,c])

def test_qarray_basic_math_tensor():
    N = 3
    a = jqt.displace(N,1.0)
    b = jqt.displace(N,1.25)
    c = jqt.displace(5,1.5)
    
    arr = jqt.Qarray.from_list([a,b])

    # ----
    tarr = jnp.array([[1,2,3,4],[4,5,6,7.2]])
    tqarr = jqt.displace(N,tarr)
    
    # Array ^ Qarray
    with pytest.raises(TypeError):
        tarr ^ b
    # assert jnp.max(jnp.abs((a.data ^ b).data - (jnp.kron(a.data,b.data)))) < 1e-10

    # Qarray ^ Array
    with pytest.raises(TypeError):
        a ^ tarr
    # assert jnp.max(jnp.abs((a ^ b.data).data - (jnp.kron(a.data,b.data))) < 1e-10)

    # Array ^ QarrayArray
    with pytest.raises(TypeError):
        tarr ^ tqarr
    # assert jnp.max(jnp.abs((a.data ^ arr).data - (jnp.kron(a.data,arr.data))) < 1e-10)
        
    # QarrayArray ^ Array
    with pytest.raises(TypeError):
        tqarr ^ tarr
    # assert jnp.max(jnp.abs((arr ^ b.data).data - (jnp.kron(arr.data,b.data))) < 1e-10)

    # ----

    # QarrayArray ^ Qarray
    assert jnp.max(jnp.abs((arr ^ b).data - (jnp.kron(arr.data,b.data))) < 1e-10)

    # Qarray ^ QarrayArray
    assert jnp.max(jnp.abs((a ^ arr).data - (jnp.kron(a.data,arr.data))) < 1e-10)

    # Qarray ^ Qarray
    assert jnp.max(jnp.abs((a ^ b).data - (jnp.kron(a.data,b.data))) < 1e-10)

    # QarrayArray ^ QarrayArray
    assert jnp.max(jnp.abs((jqt.tensor(arr,arr,parallel=True)).data - jnp.einsum("nij,nkl->nijkl", arr.data, arr.data).reshape(
                arr.data.shape[0],
                arr.data.shape[1] * arr.data.shape[1],
                -1
            )) < 1e-10)

    # QarrayArray ^ QarrayArray (of different size)
    assert  jnp.max(jnp.abs((jqt.Qarray.from_list([a,b]) ^ jqt.Qarray.from_list([a,b,b])).data - 
        jqt.Qarray.from_list([a^a, a^b, a^b, b^a, b^b, b^b]).data)) < 1e-10

    # QarrayArray ^ QarrayArray (of different size)
    arr1 = jqt.Qarray.from_list([a,b])
    arr2 = jqt.Qarray.from_list([a,b,b])
    assert jnp.max(jnp.abs((arr1^arr2).data - (
        jnp.kron(arr1.data, arr2.data)
    ))) < 1e-10

def test_qarray_basic_math_pow():
    a = jqt.displace(3,1.0)
    b = jqt.displace(3,1.25)
    c = jqt.displace(5,1.5)
    
    arr = jqt.Qarray.from_list([a,b])

    scalar = 3

    # ----
    # Qarray ** scalar

    assert jnp.max(jnp.abs((a ** scalar).data - jnp.linalg.matrix_power(a.data, scalar))) < 1e-10

    # QarrayArray ** scalar
    assert jnp.max(jnp.abs((arr ** scalar).data - jnp.linalg.matrix_power(arr.data, scalar))) < 1e-10
    assert jnp.max(jnp.abs((arr ** scalar).data - jqt.Qarray.from_list([a**scalar, b**scalar]).data)) < 1e-10

# ========================================


# Advanced Math
# ========================================
def test_eigh():
    a = jqt.create(4)
    a_dag = a.dag()

    omega = jnp.array([[1.2, 3.4,5.6],[7.8, 9.10, 11.12]])

    H = omega * a_dag @ a

    evals, evecs = jqt.eigenstates(H)

    assert evecs.shape == (2,3,4,4,1)
    assert evals[0,1,1] == 3.4



# =========================================



# Dims
# ========================================
def test_dims():
    a = jqt.displace(3,1.0)

    assert str(a.qdims) == str(a.qdims.dims)
    assert a.qdims.__repr__() == str(a.qdims)

    b = jqt.displace(4,1.0)
    assert a.qdims != b.qdims

    c = {}
    c[a.qdims] = "test"

    with pytest.raises(TypeError):
        a @ b


def test_qtype():
    a = jqt.Qtypes.ket
    b = jqt.Qtypes.bra


    assert a == jqt.Qtypes.from_str("ket")
    assert a != b

    c = jqt.basis(3, 0)
    d = jqt.basis(3, 0).dag()

    assert c.is_vec()
    assert d.is_vec()
    assert not c.to_dm().is_vec()
    assert not d.to_dm().is_vec()

    assert str(a) == "ket"
    assert str(a) == a.__repr__()

    # Hash
    test = {}
    test[a] = 1

    with pytest.raises(ValueError):
        jqt.Qtypes.from_dims([[1,2],[3,4]])

 


# def test_qtypes():

# =========================================


""" Helpers. """

from typing import List
from jax import config, Array

import jax.numpy as jnp
from tqdm import tqdm

from jaxquantum.core.qarray import Qarray, powm
from jaxquantum.core.dims import Qtypes
from jaxquantum.core.operators import identity, sigmax, sigmay, sigmaz

config.update("jax_enable_x64", True)


# Calculations ----------------------------------------------------------------

def overlap(rho: Qarray, sigma: Qarray) -> Array:
    """Overlap between two states or operators.

    Args:
        rho: state/operator.
        sigma: state/operator.

    Returns:
        Overlap between rho and sigma.
    """

    if rho.is_vec() and sigma.is_vec():
        return jnp.abs(((rho.to_ket().dag() @ sigma.to_ket()).trace())) ** 2
    elif rho.is_vec():
        rho = rho.to_ket()
        res = (rho.dag() @ sigma @ rho).data
        return res.squeeze(-1).squeeze(-1)
    elif sigma.is_vec():
        sigma = sigma.to_ket()
        res = (sigma.dag() @ rho @ sigma).data
        return res.squeeze(-1).squeeze(-1)
    else:
        return (rho.dag() @ sigma).trace()


def fidelity(rho: Qarray, sigma: Qarray) -> float:
    """Fidelity between two states.

    Args:
        rho: state.
        sigma: state.

    Returns:
        Fidelity between rho and sigma.
    """
    rho = rho.to_dm()
    sigma = sigma.to_dm()

    sqrt_rho = powm(rho, 0.5)

    return ((powm(sqrt_rho @ sigma @ sqrt_rho, 0.5)).tr()) ** 2


def quantum_state_tomography(rho: Qarray, physical_basis: Qarray,
                             logical_basis: Qarray) -> Qarray:

    """Perform quantum state tomography to retrieve the density matrix in 
    the logical basis. 

        Args:
            rho: state expressed in the physical Hilbert space basis.
            physical_basis: list of logical operators expressed in the physical
            Hilbert space basis forming a complete logical operator basis.
            logical_basis: list of logical operators expressed in the 
            logical Hilbert space basis forming a complete operator basis.


        Returns:
            Density matrix of state rho expressed in the logical basis.
        """
    dm = jnp.zeros_like(logical_basis[0].data)
    rho = rho.to_dm()

    if physical_basis.bdims[-1]!=logical_basis.bdims[-1]:
        raise ValueError(f"The two bases should have the same size for the "
                         f"last batch dimension. Received "
                         f"{physical_basis.bdims} and {logical_basis.bdims} "
                         f"instead.")

    space_size = physical_basis.bdims[-1]

    for i in tqdm(range(space_size), total=space_size):
        p_i = (rho @ physical_basis[i]).trace()
        dm += p_i * logical_basis[i].data

    return Qarray.create(dm, dims=logical_basis.dims, bdims=physical_basis[0].bdims)


def get_physical_basis(qubits: List) -> Qarray:
    """Compute a complete operator basis of a QEC code on a
    physical system specified by a number of qubits.

            Args:
                qubits: list of qubit codes, must have
                common_gates and params attributes.

            Returns:
                List containing the complete operator basis.
            """
    
    qubit = qubits[0]
    qubits = qubits[1:]
    try:
        operators = Qarray.from_list([identity(qubit.params["N"]),
                                      qubit.common_gates["X"],
                                      qubit.common_gates["Y"],
                                      qubit.common_gates["Z"]])
    except KeyError:
        print("QEC code must have common_gates for all three axes.")
    except AttributeError:
        print("QEC code must have common_gates and params attribute.")

    if len(qubits) == 0:
        return operators

    sub_basis = get_physical_basis(qubits)
    basis = []

    sub_basis_size = sub_basis.bdims[-1]

    for i in range(4):
        for j in range(sub_basis_size):
            basis.append(operators[i] ^ sub_basis[j])

    return Qarray.from_list(basis)


def get_logical_basis(n_qubits: int) -> Qarray:
    """Compute a complete operator basis of a system composed of logical
    qubits.

                Args:
                    n_qubits: number of qubits

                Returns:
                    List containing the complete operator basis.
                """
    if n_qubits < 1:
        raise ValueError("n_qubits must be at least 1.")

    n_qubits -= 1

    operators = Qarray.from_list([identity(2) / 2, sigmax() / 2, sigmay() /
                                  2, sigmaz() / 2])

    if n_qubits == 0:
        return operators

    sub_basis = get_logical_basis(n_qubits)
    basis = []

    sub_basis_size = sub_basis.bdims[-1]

    for i in range(4):
        for j in range(sub_basis_size):
            basis.append(operators[i] ^ sub_basis[j])

    return Qarray.from_list(basis)
